/*
 *  tileofday.c:	Tile of the day support
 *
 *  Written by:		Ullrich Hafner
 *  
 *  Copyright (C) 1998 Ullrich Hafner <hafner@bigfoot.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 */

/*
 *  $Date: 1999/04/27 16:08:30 $
 *  $Author: hafner $
 *  $Revision: 1.60 $
 *  $State: Exp $
 */

#include "config.h"

#include <gtk/gtk.h>
#include <proplist.h>
#include <stdio.h>
#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <sys/types.h>
#include <sys/wait.h>

#include "misc.h"
#include "dialog.h"
#include "window.h"
#include "texture.h"
#include "tileofday.h"
#include "icons.h"
#include "menu.h"

#include "error.h"

#ifdef TILE_OF_DAY

/*******************************************************************************

			global variables 
  
*******************************************************************************/

extern char		 *wmpath;	/* wmconfig.c */
extern bool_t		 changed;	/* window.c */
extern bool_t		 use_imlib;	/* window.c */
extern GtkTooltips	 *tooltips;	/* window.c */
extern proplist_t	 *windowmaker;	/* window.c */

/*******************************************************************************

		          local variables
  
*******************************************************************************/

typedef enum {TILE, BAR_FTITLE, BAR_PTITLE, BAR_UTITLE, BAR_MTITLE,
	      BAR_MTEXT, LAST_PIXMAP} tiles_e;
static char **names = NULL;

#if defined(HAVE_POPEN) && defined(CRONTAB)
static char	**cron_table = NULL;
static int	cron_offset [LAST_PIXMAP];
static char	*cron_minute [LAST_PIXMAP];
static char	*cron_hour [LAST_PIXMAP];
static char	*cron_day [LAST_PIXMAP];
static bool_t	cron_set [LAST_PIXMAP];
#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */

static char	*tile_string = "tile-of-the-day";

static GtkWidget *pixmap_tileofday [3];
static GtkWidget *bar_pixmap;

#if defined(HAVE_POPEN) && defined(CRONTAB)
static GtkWidget *combo [3];
static GtkWidget *toggle_button;
static bool_t    cron_changed = NO;
#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */

/*******************************************************************************

				prototypes
  
*******************************************************************************/

static void
cancel_http_request (GtkWidget *widget, gpointer ptr);
static void
get_tileofday (GtkWidget *widget, gpointer ptr);
static void
set_tileofday (GtkWidget *widget, gpointer ptr);
static void
cleanup (GtkWidget *widget, gpointer ptr);
static char *
bar_of_day_name (tiles_e type);
static void
init_tileofday (void);
static bool_t
generate_pixmap_templates (void);
#if defined(HAVE_POPEN) && defined(CRONTAB)
static void
set_combo (GtkWidget *widget, gpointer ptr);
static void
set_toggle (GtkWidget *widget, gpointer ptr);
static void
set_default_time (GtkWidget *widget, gpointer ptr);
static void
set_offset (GtkWidget *widget, gpointer ptr);
static void
remove_line (GtkWidget *widget, gpointer ptr);
static void
tile_toggle (GtkWidget *widget, gpointer ptr);
#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */

/*******************************************************************************

				public code
  
*******************************************************************************/

void
tileofday_dialog (GtkWidget *widget, gpointer ptr)
/*
 *  Dialog window for tile of the day configuration
 *
 *  No return value.
 */
{
   wtexture_t	*wtexture = (wtexture_t *) ptr;
   char		**tile;
   tiles_e	type;
   GtkWidget	*window;
   GtkWidget	*close_button;
   GtkWidget	*box, *button;
   GSList	*group = NULL;

   if (!names)
      init_tileofday ();
   
   for (type = 0; type < LAST_PIXMAP; type++)
      if (streq (PLGetString (wtexture->key), names [type]))
	 break;
   if (type == LAST_PIXMAP)
      return;

   if (!generate_pixmap_templates ())
      return;

   window = gtk_dialog_new ();
   gtk_window_set_title (GTK_WINDOW (window), "http://themes.org");
   gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);

   box = gtk_hbox_new (FALSE, 5);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->vbox),
		       box, FALSE, FALSE, 10);

#ifdef HAVE_IMLIB
   if (use_imlib)
   {
      GtkWidget *frame;
      
      frame = gtk_frame_new (NULL);
      gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
      gtk_container_add (GTK_CONTAINER (frame),
			 make_pixmap (PKGDATADIR "/themes.org.jpg", -1, -1,
				      NULL));
      gtk_box_pack_start (GTK_BOX (box), frame, TRUE, FALSE, 10);
   }
#elif defined(PREVIEWS) && !defined(CONVERT)
   {
      GtkWidget *frame;
      
      frame = gtk_frame_new (NULL);
      gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
      gtk_container_add (GTK_CONTAINER (frame),
			 make_pixmap (PKGDATADIR "/themes.org.jpg", -1, -1,
				      NULL));
      gtk_box_pack_start (GTK_BOX (box), frame, TRUE, FALSE, 10);
   }
#endif  /* not HAVE_IMLIB */

   box = gtk_hbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->vbox),
		       box, TRUE, FALSE, 5);
	 
   if (type == TILE)
   {
      char *tiles [] = {"tile-of-the-day", "random-tile", "guest-tile", NULL};

      for (tile = tiles; *tile; tile++)
      {
	 GtkWidget *vbox, *hbox, *button, *frame;

	 vbox = gtk_vbox_new (FALSE, 0);
	 gtk_box_pack_start (GTK_BOX (box), vbox, TRUE, FALSE, 5);

	 hbox = gtk_hbox_new (FALSE, 0);
	 gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 5);
	 frame = gtk_frame_new (NULL);
	 gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
	 pixmap_tileofday [tile - tiles] = make_pixmap (*tile, 64, 64, NULL);
	 gtk_container_add (GTK_CONTAINER (frame),
			    pixmap_tileofday [tile - tiles]);
	 gtk_box_pack_start (GTK_BOX (hbox), frame, TRUE, FALSE, 5);

	 button = gtk_radio_button_new_with_label (group, *tile);
	 gtk_object_set_user_data (GTK_OBJECT (button), (gpointer) wtexture);
	 group  = gtk_radio_button_group (GTK_RADIO_BUTTON (button));
	 gtk_box_pack_start (GTK_BOX (vbox), button, TRUE, TRUE, 5);
	 gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
				       streq (tile_string, *tile));
#if defined(HAVE_POPEN) && defined(CRONTAB)
	 gtk_signal_connect (GTK_OBJECT (button), "toggled",
			     GTK_SIGNAL_FUNC (tile_toggle), *tile);
#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */
      }
   }
   else
   {
      GtkWidget *frame;
	 
      frame = gtk_frame_new (NULL);
      bar_pixmap = make_pixmap (bar_of_day_name (type), 300, 22, NULL);
      gtk_container_add (GTK_CONTAINER (frame), bar_pixmap);
      gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
      gtk_box_pack_start (GTK_BOX (box), frame, TRUE, FALSE, 50);
   }
      
#if defined(HAVE_POPEN) && defined(CRONTAB)
   {
      GtkWidget *hbox;
      GtkWidget *vbox2;
      GtkWidget *button;
      GtkWidget *label;

      box = GTK_DIALOG (window)->vbox;
      gtk_box_pack_start (GTK_BOX (box), gtk_hseparator_new (),
			  FALSE, TRUE, 0);
	 
      hbox = gtk_hbox_new (FALSE, 10);
      gtk_box_pack_start (GTK_BOX (box), hbox, FALSE, TRUE, 10);
      
      vbox2 = gtk_vbox_new (FALSE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), vbox2, FALSE, TRUE, 5);
      toggle_button 
	 = gtk_check_button_new_with_label (_("Periodically get pixmap at"));
      button = toggle_button;
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
				    cron_set [type]);
      gtk_signal_connect (GTK_OBJECT (button), "toggled",
			  GTK_SIGNAL_FUNC (set_toggle), (gpointer) wtexture);
      gtk_box_pack_start (GTK_BOX (vbox2), button, FALSE, TRUE, 5);
      gtk_tooltips_set_tip (tooltips, button,
			    _("Add/remove entry in your crontab file to"
			      " automatically get pixmap"
			      " at specified time and day."), NULL);
      /*
       *  Specify a offset for bar of the day
       */
      if (type != TILE)
      {
	 GtkWidget *hbox = gtk_hbox_new (FALSE, 0);
	 GtkObject *adj;
	 GtkWidget *spinner;

	 gtk_box_pack_start (GTK_BOX (vbox2), hbox, FALSE, TRUE, 5);

	 gtk_box_pack_start (GTK_BOX (hbox),
			     gtk_label_new (_("Date offset [days]")),
			     FALSE, TRUE, 5);
	    
	 adj = gtk_adjustment_new (cron_offset [type], -50, 0, 1.0, 5.0, 0.0);
	 spinner = gtk_spin_button_new (GTK_ADJUSTMENT (adj), 0, 0);
	 gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinner), TRUE);
	 gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
	 gtk_widget_show_all (spinner);
	 gtk_object_set_user_data (GTK_OBJECT (adj), (gpointer) spinner);
	 gtk_tooltips_set_tip (tooltips, spinner,
			       _("When getting `bars' on a daily basis, "
				 "you can specify an offset in days "
				 "(today - offset). So you can get "
				 "different `bars' for different attributes."),
			       NULL);
	 gtk_box_pack_start (GTK_BOX (hbox), spinner, TRUE, TRUE, 0);
	 gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
			     GTK_SIGNAL_FUNC (set_offset),
			     (gpointer) &cron_set [type]);
	    
	 gtk_widget_show_all (hbox);
      }
	 
      vbox2 = gtk_vbox_new (FALSE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), vbox2, FALSE, TRUE, 5);
      label = gtk_label_new (_("Hour:"));
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, TRUE, 0);

      {
	 GList	*cbitems = NULL;
	 GtkWidget	*cb;
    
	 cbitems = g_list_append (cbitems, "0");
	 cbitems = g_list_append (cbitems, "4");
	 cbitems = g_list_append (cbitems, "8");
	 cbitems = g_list_append (cbitems, "12");
	 cbitems = g_list_append (cbitems, "16");
	 cbitems = g_list_append (cbitems, "20");
	 cbitems = g_list_append (cbitems, "*");
	    
	 combo [0] = cb = gtk_combo_new ();
	 gtk_tooltips_set_tip (tooltips, GTK_COMBO(cb)->entry,
			       _("* matches every hour"), NULL);  
	 gtk_combo_set_popdown_strings (GTK_COMBO (cb), cbitems);
	 gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(cb)->entry),
			     cron_hour [type]);
	 gtk_widget_set_usize (GTK_COMBO (cb)->entry, 80, 0);
	 gtk_widget_set_usize (cb, 80, 0);
	 gtk_box_pack_start (GTK_BOX (vbox2), cb, FALSE, TRUE, 0);
	 gtk_signal_connect (GTK_OBJECT (GTK_COMBO(cb)->entry), "changed",
			     GTK_SIGNAL_FUNC (set_combo), &cron_hour [type]);
	 gtk_object_set_user_data (GTK_OBJECT (GTK_COMBO(cb)->entry),
				   &cron_set [type]);
	 g_list_free (cbitems);
      }

      vbox2 = gtk_vbox_new (FALSE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), vbox2, FALSE, TRUE, 5);
      label = gtk_label_new (_("Minute:"));
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, TRUE, 0);

      {
	 GList	*cbitems = NULL;
	 GtkWidget	*cb;
    
	 cbitems = g_list_append (cbitems, "0");
	 cbitems = g_list_append (cbitems, "10");
	 cbitems = g_list_append (cbitems, "20");
	 cbitems = g_list_append (cbitems, "30");
	 cbitems = g_list_append (cbitems, "40");
	 cbitems = g_list_append (cbitems, "50");
	    
	 combo [1] = cb = gtk_combo_new ();
	 gtk_combo_set_popdown_strings (GTK_COMBO (cb), cbitems);
	 gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(cb)->entry),
			     cron_minute [type]);
	 gtk_widget_set_usize (GTK_COMBO (cb)->entry, 80, 0);
	 gtk_widget_set_usize (cb, 80, 0);
	 gtk_box_pack_start (GTK_BOX (vbox2), cb, FALSE, TRUE, 0);
	 gtk_signal_connect (GTK_OBJECT (GTK_COMBO(cb)->entry), "changed",
			     GTK_SIGNAL_FUNC (set_combo), &cron_minute [type]);
	 gtk_object_set_user_data (GTK_OBJECT (GTK_COMBO(cb)->entry),
				   &cron_set [type]);
	 g_list_free (cbitems);
      }
      {
	 GtkWidget *button;

	 button = gtk_button_new_with_label (_("Default"));
	 gtk_box_pack_start (GTK_BOX (vbox2), button, FALSE, TRUE, 10);
	 gtk_signal_connect (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC (set_default_time), NULL);
	 gtk_object_set_user_data (GTK_OBJECT (button),
				   type == TILE ? "tile" :"no");
	 gtk_tooltips_set_tip (tooltips, button,
			       _("Set a default time for cronjob based on "
				 "the selected tile."), NULL);  
      }
	 
	 
      vbox2 = gtk_vbox_new (FALSE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), vbox2, FALSE, TRUE, 5);
      label = gtk_label_new (_("Day of week:"));
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, TRUE, 0);

      {
	 GList *cbitems = NULL;
	 GtkWidget *cb;
	    
	 cbitems = g_list_append (cbitems, "*");
	 cbitems = g_list_append (cbitems, "Sun");
	 cbitems = g_list_append (cbitems, "Mon");
	 cbitems = g_list_append (cbitems, "Tue");
	 cbitems = g_list_append (cbitems, "Wed");
	 cbitems = g_list_append (cbitems, "Thu");
	 cbitems = g_list_append (cbitems, "Fri");
	 cbitems = g_list_append (cbitems, "Sat");
	    
	 combo [2] = cb = gtk_combo_new ();
	 gtk_tooltips_set_tip (tooltips, GTK_COMBO(cb)->entry,
			       _("* matches every day"), NULL);  
	 gtk_combo_set_popdown_strings (GTK_COMBO (cb), cbitems);
	 gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(cb)->entry),
			     cron_day [type]);
	 gtk_widget_set_usize (GTK_COMBO (cb)->entry, 80, 0);
	 gtk_widget_set_usize (cb, 80, 0);
	 gtk_box_pack_start (GTK_BOX (vbox2), cb, FALSE, TRUE, 0);
	 gtk_signal_connect (GTK_OBJECT (GTK_COMBO(cb)->entry), "changed",
			     GTK_SIGNAL_FUNC (set_combo), &cron_day [type]);
	 gtk_object_set_user_data (GTK_OBJECT (GTK_COMBO(cb)->entry),
				   &cron_set [type]);
	 g_list_free (cbitems);
      }
   }
#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */
      
   box = gtk_hbutton_box_new ();
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->action_area),
		       box, FALSE, TRUE, 5);

   button = gtk_button_new_with_label (_("Use current pixmap"));
   gtk_box_pack_start (GTK_BOX (box), button, FALSE, TRUE, 5);
   gtk_object_set_user_data (GTK_OBJECT (button), button);
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
		       GTK_SIGNAL_FUNC (set_tileofday), (gpointer) wtexture);
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			      GTK_SIGNAL_FUNC (gtk_widget_destroy),
			      GTK_OBJECT (window));

   button = gtk_button_new_with_label (_("Get new pixmap"));
   gtk_box_pack_start (GTK_BOX (box), button, FALSE, TRUE, 5);
   gtk_object_set_user_data (GTK_OBJECT (button), button);
   gtk_tooltips_set_tip (tooltips, button,
			 _("Get new pixmap via HTTP request - "
			   "this may take some time ..."), NULL);  
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
		       GTK_SIGNAL_FUNC (get_tileofday), (gpointer) wtexture);

   close_button = gtk_button_new_with_label (_("Close"));
   gtk_box_pack_start (GTK_BOX (box), close_button, FALSE, TRUE, 5);
   gtk_signal_connect (GTK_OBJECT (close_button), "clicked",
		       GTK_SIGNAL_FUNC (cleanup), NULL);
   gtk_signal_connect_object (GTK_OBJECT (close_button), "clicked",
			      GTK_SIGNAL_FUNC (gtk_widget_destroy),
			      GTK_OBJECT (window));
   gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		       GTK_SIGNAL_FUNC (delete_button), close_button);
   
   gtk_widget_show_all (window);
}

void
cleanup_tileofday (void)
{
#if defined(HAVE_POPEN) && defined(CRONTAB)
   {
      tiles_e type;

      for (type = 0; type < LAST_PIXMAP; type++)
      {
	 Free (cron_day [type]);
	 cron_day [type] = NULL;
	 Free (cron_minute [type]);
	 cron_minute [type] = NULL;
	 Free (cron_hour [type]);
	 cron_hour [type] = NULL;
      }
   }
   Free (cron_table);
   cron_table = NULL;
#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */
   Free (names);
   names = NULL;
}

#if defined(HAVE_POPEN) && defined(CRONTAB)
void
check_valid_tileofday (wtexture_t *wtexture)
/*
 *  IconBack must be 'tile-of-the-day' if cron job is activated.
 *
 *  No return value.
 */
{
   tiles_e type;

   if (!names)
      init_tileofday ();

   for (type = 0; type < LAST_PIXMAP; type++)
      if (strcaseeq (PLGetString (wtexture->key), names [type])
	  && cron_set [type])
      {
	 char	    *txt;
	 proplist_t *array = PLGetDictionaryEntry (windowmaker, wtexture->key);

	 txt = PLGetString (PLGetArrayElement (array, 0));

	 if (!strcaseeq (txt + 1, "pixmap"))
	 {
	    dialog_popup (DIALOG_WARNING, NULL, NULL,
			  _("Attribute %s:\nCan't use texture type `%s'\n"
			    "with http://themes.org cronjob\n"
			    "Removing crontab entry ..."), names [type], txt);
	    cron_set [type] = NO;
	    toggle_save (cron_changed = YES, NULL);
	    if (toggle_button)
	       gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle_button),
					    NO);
	 }
	 else
	 {
	    txt = gtk_entry_get_text (GTK_ENTRY(wtexture->wpixmap [PIXMAP_NAME]));
	    if ((type == TILE && !streq (txt, tile_string))
		|| (type != TILE && !streq (txt, bar_of_day_name (type))))
	    {
	       char *pname;

	       if (type == TILE)
		  pname = tile_string;
	       else
		  pname = bar_of_day_name (type);
	       dialog_popup (DIALOG_WARNING, NULL, NULL,
			     _("Attribute %s: pixmap value of\n"
			       "cronjob [%s] doesn't match value\n"
			       "of WindowMaker file [%s].\n"
			       "Removing crontab entry ..."),
			     names [type], pname, txt);
	       cron_set [type] = NO;
	       toggle_save (cron_changed = YES, NULL);
	       if (toggle_button)
		  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle_button),
					       NO);
	    }
	 }
      }
}

void
read_crontab (void)
/*
 *  Read cron table and set local variables.
 *
 *  No return value.
 *
 *  Side effects:
 *	locals 'cron_day', 'crontab_file', 'cron_set', 'cron_minute',
 *	'cron_hour' are set.
 */
{
   FILE		*cronpipe;
   tiles_e	type;

   if (!names)
      init_tileofday ();

   cronpipe = popen (CRONTAB " -l", "r");
   if (!cronpipe)
   {
      dialog_popup (DIALOG_ERROR, NULL, NULL,
		    _("Can't connect to crontab utility."));
      return;
   }
   else
   {
      unsigned	lines = 0;		/* number of lines in crontab */
      char	tmp [MAXSTRLEN];

      while (fgets (tmp, MAXSTRLEN, cronpipe))
	 lines++;
      pclose (cronpipe);

      cron_table = Calloc (lines + 1, sizeof (char *));
      if (lines > 0)
      {
	 /*
	  *  Read and store lines of crontab
	  */
	 cronpipe = popen (CRONTAB " -l", "r");
	 if (!cronpipe)
	 {
	    dialog_popup (DIALOG_ERROR, NULL, NULL,
			  _("Can't connect to crontab utility."));
	    return;
	 }
	 else
	 {
	    lines = 0;
	    while (fgets (tmp, MAXSTRLEN, cronpipe))
	    {
	       if (*tmp == '#')
		  continue;
	       if (strstr (tmp, "gettile.pl")) /* found script call */
	       {
		  char	dummy [11][MAXSTRLEN];
		  int	values;
		  bool_t	conversion = NO;
	       
		  for (type = 0; type < LAST_PIXMAP; type++)
		     if (strstr (tmp, names [type]))
			break;
		  if (type == LAST_PIXMAP)	/* maybe wmakerconf < 1.3 */
		  {
		     conversion = YES;
		     type	     = TILE;
		  }
	       
		  Free (cron_minute [type]);
		  Free (cron_hour [type]);
		  Free (cron_day [type]);

		  if ((values = sscanf (tmp, "%s %s %s %s %s %s %s %s %s %s %s",
					dummy [0], dummy [1], dummy [2],
					dummy [3], dummy [4], dummy [5],
					dummy [6], dummy [7], dummy [8],
					dummy [9], dummy [10])) < 10)
		  {
		     if (values != 9 || !conversion) /* crontable syntax error */
		     {
			cron_table [lines] = g_strdup (tmp);
			dialog (DIALOG_QUESTION, _("No, keep this line"),
				_("Yes, remove this line"), remove_line,
				cron_table [lines],
				_("Crontable entry\n"
				   "%scalls gettile.pl but doesn't"
				   " contain valid entry with 10 fields\n\n"
				   "Should I delete this line "
				   "from your crontable?"), tmp);
			lines++;
			continue;
		     }
		  }
		  else if (values == 10 && type != TILE)
		  {
		     cron_table [lines] = g_strdup (tmp);
		     dialog (DIALOG_QUESTION, _("No, keep this line"),
			     _("Yes, remove this line"), remove_line,
			     cron_table [lines],
			     _("Crontable entry\n%s"
				"\ncalls gettile.pl but doesn't"
				" contain valid entry with 11 fields\n\n"
				"Should I delete this line from your crontable?"),
			      tmp);
		     lines++;
		     continue;
		  }
	       
		  cron_minute [type] = g_strdup (dummy [0]);
		  cron_hour [type]   = g_strdup (dummy [1]);
		  cron_day [type]    = g_strdup (dummy [4]);
		  cron_set [type]    = YES;
		  if (type == TILE)
		  {
		     if (strstr (tmp, "tile-of-the-day"))
			tile_string = "tile-of-the-day";
		     else if (strstr (tmp, "random-tile"))
			tile_string = "random-tile";
		     else
			tile_string = "guest-tile";
		  }
		  else 
		     cron_offset [type] = strtod (dummy [10], NULL);
	       }
	       else
		  cron_table [lines++] = g_strdup (tmp);
	    }
	    pclose (cronpipe);
	 }
      }
      cron_table [lines] = NULL;
   }
}

bool_t
save_crontab (void)
/*
 *  Save new cron table if required, i.e. cron_changed == TRUE.
 *
 *  Return value:
 *	TRUE if file has been saved
 */
{
   if (cron_changed)
   {
      tiles_e	type;
      char	*tmp_name      = tmpnam (NULL);
      char	*cron_filename = tmp_name ? tmp_name : "/tmp/cron.tmp";
      FILE	*cronfile      = fopen (cron_filename, "w");
      
      if (!cronfile)
      {
	 dialog_popup (DIALOG_ERROR, NULL, NULL,
		       _("Can't write to temporary crontab file."));
	 return FALSE;
      }
      if (cron_table)			/* write all non-gettile entries */
      {
	 char **line;
	 
	 for (line = cron_table; *line; line++)
	    fputs (*line, cronfile);
      }
      
      for (type = 0; type < LAST_PIXMAP; type++) /* write all gettile entries */
      {
	 char *path;

	 if (type == TILE)
	    path = get_pixmap_path (tile_string);
	 else
	    path = get_pixmap_path (bar_of_day_name (type));
	 
	 if (path && cron_set [type])
	 {
	    fprintf (cronfile, "%s %s * * %s %s %s %s %s",
		     cron_minute [type], cron_hour [type], cron_day [type],
		     PERL " " PKGDATADIR "/gettile.pl", wmpath, path,
		     names [type]);
	    if (type == TILE)
	       fprintf (cronfile, "\n");
	    else
	       fprintf (cronfile, " %d\n", cron_offset [type]);
	    Free (path);
	 }
	 else if (cron_set [type])
	    dialog_popup (DIALOG_ERROR, NULL, NULL,
			  _("Can't find pixmap\n`%s' in PixmapPath\n"
			    "(requested by attribute '%s`)."),
			  path, names [type]);
      }
      fclose (cronfile);
      {
	 char *tmp = g_strconcat (CRONTAB, " ", cron_filename, NULL);
	 if (system (tmp))
	 {
	    dialog_popup (DIALOG_ERROR, NULL, NULL,
			  _("Can't replace crontab file."));
	    Free (tmp);
	    delete_file_or_dir (tmp_name);
	    return FALSE;
	 }
	 else
	    message (_("Crontab file changed."));
	 Free (tmp);
	 delete_file_or_dir (tmp_name);
      }
   }
   cron_changed = NO;
   return TRUE;
}

bool_t
crontab_changed (void)
{
   return cron_changed;
}
#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */

/*******************************************************************************

				private code
  
*******************************************************************************/

static bool_t
generate_pixmap_templates (void)
/*
 *  Generate empty pixmap files in users pixmap directory if not already done.
 */
{
   const char *all_tiles [LAST_PIXMAP + 4];
   const char **tile;
   unsigned   pos = 0;
   tiles_e    type;

   all_tiles [pos++] = "tile-of-the-day";
   all_tiles [pos++] = "random-tile";
   all_tiles [pos++] = "guest-tile";
   
   for (type = 0; type < LAST_PIXMAP; type++)
      if (type != TILE)
	 all_tiles [pos++] = bar_of_day_name (type);
   all_tiles [pos++] =  NULL;
   
   for (tile = all_tiles; *tile; tile++)
   {
      char *path = get_pixmap_path (*tile);

      if (path)
      {
	 Free (path);
	 continue;
      }
      else
      {
	 int	  n;
	 bool_t	  success = NO;
	 char	  *home = expand_tilde ("~/GNUstep/Library/WindowMaker/Pixmaps");
	 proplist_t	*plpath      = PLMakeString ("PixmapPath");
	 proplist_t	*pixmap_path = PLGetDictionaryEntry (windowmaker, plpath);

	 PLRelease (plpath);

	 for (n = 0; n < (int) PLGetNumberOfElements (pixmap_path) && !success;
	      n++)
	 {
	    char *path = PLGetString (PLGetArrayElement (pixmap_path, n));
	    
	    path = expand_tilde (path);
	    if (strcaseeq (path, home))
	    {
	       success = YES;
	       Free (path);
	       break;
	    }
	    Free (path);
	 }

	 n       = success ? -1 : 0;
	 success = NO;
	 
	 /*
	  *  Generate an empty pixmap file 'tile'
	  */
	 for (; n < (int) PLGetNumberOfElements (pixmap_path) && !success; n++)
	 {
	    char *path;
	    FILE *dst;
	    char *tmp;

	    if (n < 0)
	       path = g_strdup (home);
	    else
	    {
	       path = PLGetString (PLGetArrayElement (pixmap_path, n));
	       path = expand_tilde (path);
	    }
	    tmp = g_strconcat (path, "/", *tile, NULL);
	    dst = fopen (tmp, "w");
	    message ("Trying to write %s ... ", tmp);
	    Free (path);
	    Free (tmp);
	    if (dst)
	    {
	       char	*buffer [256];
	       FILE	*src = fopen (PKGDATADIR "/black.xpm", "r");
	       size_t	n;
	       
	       if (!src)
		  error ("File '%s' not found. Please check installation.",
			 PKGDATADIR "/black.xpm", PKGDATADIR "/black.xpm");

	       while ((n = fread (buffer, 1, 256, src)))
		  fwrite (buffer, 1, n, dst);
	       fclose (dst);
	       fclose (src);
	       success = YES;
	       message ("done.");
	    }
	    else
	       message ("error.");
	 }
	 Free (home);
	 if (!success)
	 {
	    dialog_popup (DIALOG_ERROR, NULL, NULL,
			  _("Can't generate tile template `%s'\n"
			    "in neither of your PixmapPath directories.\n"
			    "Please ensure that you have write access\n"
			    "for at least one directory."), *tile);
	    return FALSE;
	 }
      }
   }
   return TRUE;
}

static void
set_tileofday (GtkWidget *widget, gpointer ptr)
/*
 *  Set pixmap of IconBack to 'tile-of-the-day'.
 *  Update corresponding widgets accordingly.
 *
 *  No return value.
 */
{
   wtexture_t	*wtexture = (wtexture_t *) ptr;
   char		*tilename;
   char		*path;
   tiles_e	type;
   
   for (type = 0; type < LAST_PIXMAP; type++)
      if (strcaseeq (PLGetString (wtexture->key), names [type]))
	 break;

   if (strcaseeq (PLGetString (wtexture->key), names [TILE]))
      tilename = tile_string;
   else
      tilename = bar_of_day_name (type);
   path = get_pixmap_path (tilename);
   
   if (path)
   { 
      proplist_t *array = PLGetDictionaryEntry (windowmaker, wtexture->key);
      
      Free (path);
      PLRemoveArrayElement (array, 1);	/* filename */
      PLInsertArrayElement (array, PLMakeString (tilename), 1);

      show_pixmap (wtexture->key, tilename, wtexture->wpixmap [PIXMAP_PIXMAP]);
      gtk_entry_set_text (GTK_ENTRY (wtexture->wpixmap [PIXMAP_NAME]), tilename);
      toggle_save (changed = YES, wtexture->key);
   }
   else
      dialog_popup (DIALOG_ERROR, NULL, NULL,
		    _("Can't find pixmap\n`%s' in PixmapPath\n"
		      "(requested by attribute '%s`)."),
		    tilename, PLGetString (wtexture->key));
}

static void
get_tileofday (GtkWidget *widget, gpointer ptr)
/*
 *  Call perl script gettile.pl that tries to get themes.org current
 *  tile of the day.
 *
 *  No return value.
 */
{
   wtexture_t	*wtexture = (wtexture_t *) ptr;
   char		*tilename;
   char		*path;
   tiles_e	type;
   
   for (type = 0; type < LAST_PIXMAP; type++)
      if (strcaseeq (PLGetString (wtexture->key), names [type]))
	 break;

   if (strcaseeq (PLGetString (wtexture->key), names [TILE]))
      tilename = tile_string;
   else
      tilename = bar_of_day_name (type);
    path = get_pixmap_path (tilename);

   if (path && type < LAST_PIXMAP)
   {
      pid_t	pid;
      GtkWidget *progress_window;
      GtkWidget *progress_label;
      GtkWidget *progress_bar;
      GtkWidget *vbox;
      gint	timer;
   
      /*
       *  Generate a progressbar and window
       */
      progress_window = gtk_dialog_new ();
      gtk_window_set_position (GTK_WINDOW (progress_window), GTK_WIN_POS_MOUSE);
      gtk_window_set_title (GTK_WINDOW (progress_window), _("Get new pixmap"));
      gtk_widget_set_usize (progress_window, 250, -1);
   
      vbox = gtk_vbox_new (FALSE, 5);
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (progress_window)->vbox),
			  vbox, TRUE, TRUE, 0);
      gtk_container_set_border_width (GTK_CONTAINER (vbox), 10);

      progress_label = gtk_label_new ("");
      gtk_box_pack_start (GTK_BOX (vbox), progress_label, FALSE, TRUE, 0);
      if (type == TILE)
	 gtk_label_set_text (GTK_LABEL (progress_label), tile_string);
      else
	 gtk_label_set_text (GTK_LABEL (progress_label), bar_of_day_name (type));
      {
	 GtkAdjustment *adj;
	 
	 adj = (GtkAdjustment *) gtk_adjustment_new (0, 1, 300, 0, 0, 0);
	 
	 progress_bar = gtk_progress_bar_new_with_adjustment (adj);
	 gtk_progress_set_activity_mode (GTK_PROGRESS (progress_bar), TRUE);
	 gtk_progress_bar_set_activity_step (GTK_PROGRESS_BAR (progress_bar), 5);
	 gtk_progress_bar_set_activity_blocks (GTK_PROGRESS_BAR (progress_bar),
					       4);
	 timer = gtk_timeout_add (100, progress_timer, progress_bar);
      }

      gtk_box_pack_start (GTK_BOX (vbox), progress_bar, FALSE, TRUE, 5);

      {
	 GtkWidget *bbox   = gtk_hbutton_box_new ();
	 GtkWidget *cancel = gtk_button_new_with_label (_("Cancel"));
	 gtk_box_pack_start (GTK_BOX (GTK_BOX (GTK_DIALOG (progress_window)->action_area)), bbox, FALSE, TRUE, 0);
	 gtk_box_pack_start (GTK_BOX (bbox), cancel, FALSE, TRUE, 0);
	 gtk_signal_connect (GTK_OBJECT (cancel), "clicked",
			     GTK_SIGNAL_FUNC (cancel_http_request), &pid);
	 gtk_signal_connect (GTK_OBJECT (progress_window), "delete_event",
			     GTK_SIGNAL_FUNC (delete_button), cancel);
      }
      
      gtk_widget_show_all (progress_window);

      pid = fork ();
      if (pid)			/* parent */
      {
	 int status;
	 
	 while (waitpid (pid, &status, WNOHANG) == 0)
	    gtk_main_iteration ();
	 if (WIFSIGNALED (status) || (WIFEXITED (status) && WEXITSTATUS (status)))
	 {
	    if (!WIFSIGNALED (status))
	       dialog_popup (DIALOG_ERROR, NULL, NULL,
			     _("Can't receive pixmap from http://themes.org/.\n"
			       "Please check `stderr' for more details."));
	    gtk_timeout_remove (timer);
	    Free (path);
	    gtk_widget_destroy (progress_window);
    
	    return;
	 }
      }
      else
      {
	 execl (PERL, PERL, PKGDATADIR "/gettile.pl", "no", path, names [type],
		NULL);
	 exit (1);
      }
      if (type == TILE)
      {
	 if (streq (tile_string, "tile-of-the-day"))
	    make_pixmap (tile_string, 64, 64, pixmap_tileofday [0]);
	 else if (streq (tile_string, "random-tile"))
	    make_pixmap (tile_string, 64, 64, pixmap_tileofday [1]);
	 else
	    make_pixmap (tile_string, 64, 64, pixmap_tileofday [2]);
      }
      else
	 make_pixmap (bar_of_day_name (type), 400, 22, bar_pixmap);
      
#if defined(HAVE_POPEN) && defined(CRONTAB)
      if (cron_set [type])
	 set_tileofday (widget, ptr);
#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */
/*       } */
      Free (path);
      gtk_timeout_remove (timer);
      gtk_widget_destroy (progress_window);
   }
   else
      dialog_popup (DIALOG_ERROR, NULL, NULL,
		    _("Can't find pixmap\n`%s' in PixmapPath\n"
		      "(requested by attribute '%s`)."),
		    tilename, PLGetString (wtexture->key));
}

static void
cancel_http_request (GtkWidget *widget, gpointer ptr)
{
   pid_t pid = * (pid_t *) ptr;

   kill (pid, 15);
}

#if defined(HAVE_POPEN) && defined(CRONTAB)
static void
set_combo (GtkWidget *widget, gpointer ptr)
{
   char **value = (char **) ptr;
   
   Free (*value);
   *value = g_strdup (gtk_entry_get_text (GTK_ENTRY (widget)));

   if (* (bool_t *) gtk_object_get_user_data (GTK_OBJECT (widget)))
      toggle_save (cron_changed = YES, NULL);
}

static void
set_toggle (GtkWidget *widget, gpointer ptr)
{
   wtexture_t	*wtexture = (wtexture_t *) ptr;
   tiles_e    type;
   
   for (type = 0; type < LAST_PIXMAP; type++)
      if (strcaseeq (PLGetString (wtexture->key), names [type]))
	 break;
   if (type == LAST_PIXMAP)
      return;

   cron_set [type] = GTK_TOGGLE_BUTTON (widget)->active;
   toggle_save (cron_changed = YES, NULL);

   if (cron_set [type])
      set_tileofday (widget, ptr);
}

static void
set_default_time (GtkWidget *widget, gpointer ptr)
{
   if (streq (tile_string, "tile-of-the-day")
       || !streq (gtk_object_get_user_data (GTK_OBJECT (widget)), "tile"))
   {
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(combo [0])->entry), "6");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(combo [1])->entry), "0");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(combo [2])->entry), "*");
   }
   else if (streq (tile_string, "random-tile"))
   {
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(combo [0])->entry), "*");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(combo [1])->entry), "1");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(combo [2])->entry), "*");
   }
   else
   {
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(combo [0])->entry), "6");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(combo [1])->entry), "0");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO(combo [2])->entry), "Tue");
   }
}

static void
remove_line (GtkWidget *widget, gpointer ptr)
{
   char **line;
   
   for (line = cron_table; *line; line++)
      if (*line == (char *) ptr)
      {
	 char **current = line;
	 
	 for (; *line; line++)
	    ;
	 line--;
	 Free (*current);
	 *current     = *line;
	 *line 	      = NULL;
	 toggle_save (cron_changed = YES, NULL);
      }
}

static void
set_offset (GtkWidget *widget, gpointer ptr)
/*
 *  Update date offset (spinner).
 *
 *  No return value.
 */
{
   GtkSpinButton *spinner;

   spinner     = GTK_SPIN_BUTTON (gtk_object_get_user_data (GTK_OBJECT (widget)));
   cron_offset [(bool_t *) ptr - cron_set]
      = gtk_spin_button_get_value_as_int (spinner);

   if (* (bool_t *) ptr)			/* crob job set */
      toggle_save (cron_changed = YES, NULL);
}

static void
tile_toggle (GtkWidget *widget, gpointer ptr)
{
   if (GTK_TOGGLE_BUTTON (widget)->active)
   {
      tile_string = (char *) ptr;
      if (cron_set [TILE])
      {
	 wtexture_t *wtexture = gtk_object_get_user_data (GTK_OBJECT (widget));
	 
	 set_tileofday (widget, wtexture);
	 toggle_save (cron_changed = YES, NULL);
      }
   }
}

#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */

static void
cleanup (GtkWidget *widget, gpointer ptr)
{
   pixmap_tileofday [0] = NULL;
   pixmap_tileofday [1] = NULL;
   pixmap_tileofday [2] = NULL;
   bar_pixmap		= NULL;
#if defined(HAVE_POPEN) && defined(CRONTAB)
   combo [0]		= NULL;
   combo [1]		= NULL;
   combo [2]		= NULL;
   toggle_button	= NULL;
#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */
}

static char *
bar_of_day_name (tiles_e type)
{
   static char **filename = NULL;

   if (type >= LAST_PIXMAP || type == TILE)
      return NULL;
   
   if (!filename)
   {
      tiles_e p;
      
      filename = Calloc (LAST_PIXMAP, sizeof (char *));
      for (p = TILE + 1; p < LAST_PIXMAP; p++)
      {
	 char tmp [MAXSTRLEN];

	 sprintf (tmp, "bar%d-of-the-day", p);
	 filename [p] = g_strdup (tmp);
      }
   }

   return filename [type];
}

static void
init_tileofday (void)
{
#if defined(HAVE_POPEN) && defined(CRONTAB)
   {
      tiles_e type;
      for (type = 0; type < LAST_PIXMAP; type++)
      {
	 cron_day [type]    = g_strdup ("*");
	 cron_minute [type] = g_strdup ("0");
	 cron_hour [type]   = g_strdup ("6");
	 cron_set [type]    = NO;
	 cron_offset [type] = 0;
      }
   }
#endif /*  defined(HAVE_POPEN) && defined(CRONTAB) */
   names 	      = Calloc (LAST_PIXMAP, sizeof (char *));
   names [TILE]       = "IconBack";
   names [BAR_FTITLE] = "FTitleBack";
   names [BAR_PTITLE] = "PTitleBack";
   names [BAR_UTITLE] = "UTitleBack";
   names [BAR_MTITLE] = "MenuTitleBack";
   names [BAR_MTEXT]  = "MenuTextBack";
}

#endif /* TILE_OF_DAY */
