/*
 *  window.c:		GTK stuff
 *
 *  Written by:		Ullrich Hafner
 *  
 *  Copyright (C) 1998 Ullrich Hafner <hafner@bigfoot.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 */


/*
 *  $Date: 1999/04/27 16:08:30 $
 *  $Author: hafner $
 *  $Revision: 1.111 $
 *  $State: Exp $
 */

#include "config.h"

#if HAVE_STDLIB_H
#	include <stdlib.h>
#endif /* not HAVE_STDLIB_H */
#if HAVE_STRING_H
#	include <string.h>
#else /* not HAVE_STRING_H */
#	include <strings.h>
#endif /* not HAVE_STRING_H */
#if HAVE_SYS_STAT_H
#	include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */

#include <stdio.h>
#include <gtk/gtk.h>
#include <proplist.h>

#ifdef HAVE_IMLIB
#include <gdk_imlib.h>
#endif /* HAVE_IMLIB */

#include "window.h"
#include "icons.h"
#include "misc.h"
#include "dialog.h"
#include "texture.h"
#include "keys.h"
#include "menu.h"
#include "tileofday.h"
#include "shortcuts.h"
#include "rootmenu.h"
#include "simple.h"
#include "path.h"
#include "themes.h"
#include "rimage.h"

#include "error.h"

/*******************************************************************************

			     global variables
  
*******************************************************************************/

bool_t		 changed      = NO;
bool_t		 use_imlib    = NO;
GtkWidget	 *main_window = NULL;
GtkTooltips	 *tooltips    = NULL;
proplist_t	 *windowmaker = NULL;
proplist_t	 *wmconfig    = NULL;

extern proplist_t *plist_changed;
extern proplist_t *pl_yes;

GtkWidget *widgetstyle_item = NULL;

/*******************************************************************************

			       prototypes
  
*******************************************************************************/

void
update (gpointer ptr, proplist_t *value);
static GtkWidget *
create_page (GtkNotebook *notebook, const char *name, GtkWidget *pixmap);
static void
quit_and_save_popup (char *question, void (*yes_function) (GtkWidget *, gpointer),
		     gpointer dataptr);
static void
switch_subentries (GtkWidget *widget, gpointer ptr);
static void
init_option_menu (gpointer data, gpointer user_data);

/*******************************************************************************

			       public code
  
*******************************************************************************/

void
root_window (GtkWidget *logwindow, proplist_t *wmc, proplist_t *panel,
	     proplist_t *wmlist)
/*
 *  Create root window and call the event loop of GTK
 *
 *  No return value.
 */
{
   GtkWidget  *box1;
   GtkWidget  *notebook;
#ifndef FREEBSD_SYSTEM   
   GtkWidget  *progress_window;
   GtkWidget  *progress_label;
   GtkWidget  *progress_bar;
#endif /* not FREEBSD_SYSTEM */
   GtkWidget  *toolbar;
   proplist_t *plinfo    = PLMakeString ("Info");
   proplist_t *pltype    = PLMakeString ("Type");
   proplist_t *plicon    = PLMakeString ("Icon");
   proplist_t *pltitle   = PLMakeString ("Title");
   proplist_t *plwidget  = PLMakeString ("Widget");
   proplist_t *plcolor   = PLMakeString ("Colorbox");
   proplist_t *plfont    = PLMakeString ("Fontbox");
   proplist_t *pltexture = PLMakeString ("Texturebox");
   proplist_t *plgrp     = PLMakeString ("Group");
   proplist_t *plrange   = PLMakeString ("Range");
   GList      *options	 = NULL;
   
#if defined(HAVE_POPEN) && defined(TILE_OF_DAY) && defined(CRONTAB)   
   read_crontab ();
#endif /* defined(HAVE_POPEN) && defined(TILE_OF_DAY) && defined(CRONTAB) */
   
   windowmaker = wmlist;
   wmconfig    = wmc;

#ifndef FREEBSD_SYSTEM   
   /*
    *  Progress bar window
    */
   {
      GtkWidget *vbox;
      
      progress_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
      gtk_window_set_title (GTK_WINDOW (progress_window),
			    PACKAGE " " VERSION " init ...");
      gtk_container_set_border_width (GTK_CONTAINER (progress_window), 10);
      gtk_signal_connect (GTK_OBJECT (progress_window), "delete_event",
			  GTK_SIGNAL_FUNC (gtk_true), NULL);
      gtk_widget_set_usize (progress_window, 250, -1);
      
      vbox = gtk_vbox_new (FALSE, 5);
      gtk_container_add (GTK_CONTAINER (progress_window), vbox);

      progress_label = gtk_label_new ("");
      gtk_box_pack_start (GTK_BOX (vbox), progress_label, FALSE, TRUE, 0);

      {
	 GtkAdjustment *adj;
	 
	 adj = (GtkAdjustment *) gtk_adjustment_new (0, 1, 300, 0, 0, 0);
	 
	 progress_bar = gtk_progress_bar_new_with_adjustment (adj);
	 gtk_progress_set_format_string (GTK_PROGRESS (progress_bar), "[%p%%]");
	 gtk_progress_set_show_text (GTK_PROGRESS (progress_bar), YES);
      }

      gtk_box_pack_start (GTK_BOX (vbox), progress_bar, FALSE, TRUE, 0);
      gtk_widget_show_all (progress_window);
   }
#endif /* not FREEBSD_SYSTEM */
   
#ifdef HAVE_IMLIB
   {
      GdkVisual		*gdkvisual;
      GdkColormap	*gdkcmap;
      
      gdk_imlib_init ();
      gdkvisual = gdk_imlib_get_visual ();
      gdkcmap   = gdk_imlib_get_colormap ();

      if (gdkvisual && gdkcmap)
      {
	 gtk_widget_push_visual (gdkvisual);
	 gtk_widget_push_colormap (gdkcmap);
	 use_imlib = YES;
      }
      else
	 use_imlib = NO;
   }
#endif /* HAVE_IMLIB */
   gtk_widget_push_visual (gdk_rgb_get_visual ());
   gtk_widget_push_colormap (gdk_rgb_get_cmap ());

   /*
    *  Main window
    */
   main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title (GTK_WINDOW (main_window), PACKAGE " " VERSION);
   gtk_signal_connect (GTK_OBJECT (main_window), "delete_event",
		       GTK_SIGNAL_FUNC (quit), NULL);
   gtk_widget_realize (main_window);

   init_pixmaps ();

   while (gtk_events_pending())
      gtk_main_iteration();

   box1 = gtk_vbox_new (FALSE, 0);
   gtk_container_add (GTK_CONTAINER (main_window), box1);
   tooltips = gtk_tooltips_new ();
   gtk_widget_show (box1);

   while (gtk_events_pending())
      gtk_main_iteration();

   /*
    *  Menu and toolbox
    */
   gtk_box_pack_start (GTK_BOX(box1), make_menubar (logwindow), FALSE, TRUE, 0);
   gtk_box_pack_start (GTK_BOX(box1), toolbar = make_toolbar (), FALSE, TRUE, 0);
   
   while (gtk_events_pending())
      gtk_main_iteration();

   /*
    *  Groups notebook
    */
   notebook = gtk_notebook_new ();
   gtk_notebook_set_show_tabs (GTK_NOTEBOOK (notebook), TRUE);
   gtk_notebook_set_scrollable (GTK_NOTEBOOK (notebook), TRUE);
   gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), GTK_POS_TOP);
   gtk_box_pack_start (GTK_BOX (box1), notebook, TRUE, TRUE, 0);
   gtk_container_set_border_width (GTK_CONTAINER (notebook), 10);
   gtk_widget_show_all (notebook);

   while (gtk_events_pending())
      gtk_main_iteration();

   /*
    *  Generate notebook page for every group
    */
   {
      proplist_t *tab = PLGetAllDictionaryKeys (panel);
      unsigned	 n;
      
      for (n = 0; n < PLGetNumberOfElements (tab); n++)
      {
	 proplist_t *thispanel = PLGetArrayElement (tab, n);
	 proplist_t *entry     = PLGetDictionaryEntry (panel, thispanel);
	 proplist_t *info      = PLGetDictionaryEntry (entry, plinfo);
	 proplist_t *icon      = PLGetDictionaryEntry (entry, plicon);
	 proplist_t *type      = PLGetDictionaryEntry (entry, pltype);
	 GtkWidget  *page;

	 if (!icon || !info || !PLIsString (info) || !PLIsString (icon))
	    error (_("'%s': parse error."), PKGDATADIR "/WMWmakerconf");
	 /*
	  *  Create new notbook page
	  */
	 {
	    GdkPixmap *pixmap;
	    GdkBitmap *mask;
	    char *path = g_strconcat (PKGDATADIR, "/", PLGetString (icon), NULL);
	    pixmap = gdk_pixmap_create_from_xpm (main_window->window, &mask,
						 &main_window->style->bg [GTK_STATE_NORMAL], path);
	    Free (path);
	    page = create_page (GTK_NOTEBOOK (notebook), D_(PLGetString (info)),
				gtk_pixmap_new (pixmap, mask));
	 }
	 while (gtk_events_pending())
	    gtk_main_iteration();
	 /*
	  *  Create container for config widgets
	  */
	 {
	    GtkWidget	*box;
	    proplist_t	*data;
	    
	    if (!PLIsString (type) || strcaseeq (PLGetString (type), "SINGLE"))
	       box = gtk_vbox_new (FALSE, 5);
	    else			/* DOUBLE */
	    {
	       unsigned *number = Calloc (1, sizeof (unsigned));
	       
	       box = gtk_table_new (1, 4, FALSE);
	       gtk_table_set_row_spacings (GTK_TABLE (box), 5);
	       gtk_table_set_col_spacings (GTK_TABLE (box), 10);
	       gtk_container_set_border_width (GTK_CONTAINER (box), 5);
	       
	       *number = 1;
	       gtk_object_set_user_data (GTK_OBJECT (box), number);
	    }
	    gtk_widget_show (box);

	    while (gtk_events_pending())
	       gtk_main_iteration();

	    /*
	     *  Frame or scrooled window for all sub widgets
	     */
	    if (strcaseeq (PLGetString (thispanel), "Shortcuts")
		|| strcaseeq (PLGetString (thispanel), "Menu")
		|| strcaseeq (PLGetString (thispanel), "Appearance"))
	    {
	       GtkWidget *frame = gtk_frame_new (NULL);
	       
	       gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
	       gtk_container_set_border_width (GTK_CONTAINER (frame), 5);
	       gtk_container_add (GTK_CONTAINER (frame), box);
	       
	       gtk_container_add (GTK_CONTAINER (page), frame);
	    }
	    else
	    {
	       GtkWidget *scrolled = gtk_scrolled_window_new (NULL, NULL);
	       
	       gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled),
					       GTK_POLICY_AUTOMATIC,
					       GTK_POLICY_AUTOMATIC);
	       gtk_container_set_border_width (GTK_CONTAINER (scrolled), 5);
	       gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolled),
						      box);

	       gtk_container_add (GTK_CONTAINER (page), scrolled);
	    }

	    gtk_widget_show_all (page);

	    while (gtk_events_pending())
	       gtk_main_iteration();

	    if (strcaseeq (PLGetString (thispanel), "Menu"))
	       rootmenu_dialog (box, tooltips);
	    if (strcaseeq (PLGetString (thispanel), "Themes"))
	       box = theme_dialog (box);

	    data = PLMakeData ((unsigned char *) &box, sizeof (GtkWidget *));
	    PLInsertDictionaryEntry (entry, plwidget, data);
	    PLRelease (data);

	    if (PLIsDictionary (type))	/* subsections available ? */
	    {
	       proplist_t *subentries = PLGetAllDictionaryKeys (type);
	       GtkWidget  *subbook    = gtk_notebook_new ();
	       unsigned	  i;
	       
	       gtk_notebook_set_show_tabs (GTK_NOTEBOOK (subbook), TRUE);
	       gtk_notebook_set_scrollable (GTK_NOTEBOOK (subbook), TRUE);
	       gtk_notebook_set_tab_pos (GTK_NOTEBOOK (subbook), GTK_POS_LEFT);
	       gtk_notebook_set_homogeneous_tabs (GTK_NOTEBOOK (subbook), TRUE);
	       gtk_box_pack_start (GTK_BOX (box), subbook, TRUE, TRUE, 0);
	       gtk_container_set_border_width (GTK_CONTAINER (subbook), 10);
	       gtk_widget_show_all (subbook);
	       
	       while (gtk_events_pending())
		  gtk_main_iteration();
	       
	       for (i = 0; i < PLGetNumberOfElements (subentries); i++)
	       {
		  proplist_t *thissub  = PLGetArrayElement (subentries, i);
		  proplist_t *subentry = PLGetDictionaryEntry (type, thissub);
		  proplist_t *subinfo  = PLGetDictionaryEntry (subentry, plinfo);
		  proplist_t *subicon  = PLGetDictionaryEntry (subentry, plicon);
		  proplist_t *subtype  = PLGetDictionaryEntry (subentry, pltype);
		  GtkWidget  *subpage;
		  GtkWidget  *picon = NULL;

		  if (subicon && PLIsString (subicon)) /* subsubsections ? */
		  {
		     GdkPixmap *pixmap;
		     GdkBitmap *mask;
		     char *path = g_strconcat (PKGDATADIR, "/",
					       PLGetString (subicon), NULL);
		     
		     pixmap = gdk_pixmap_create_from_xpm (main_window->window, &mask, &main_window->style->bg [GTK_STATE_NORMAL], path);
		     picon  = gtk_pixmap_new (pixmap, mask);
		     Free (path);
		  }
		  subpage = create_page (GTK_NOTEBOOK (subbook),
					 D_(PLGetString (subinfo)), picon);
		  data = PLMakeData ((unsigned char *) &subpage,
 				     sizeof (GtkWidget *));
		  PLInsertDictionaryEntry (subentry, plwidget, data);
		  PLRelease (data);
		  
		  while (gtk_events_pending())
		     gtk_main_iteration();
		  
		  /*
		   *  Add subboxes for color, texture and font
		   */
		  {
		     GtkWidget  *vbox          = gtk_vbox_new (FALSE, 0);
		     GtkWidget  *nw_hbox       = gtk_hbox_new (FALSE, 0);
		     GtkWidget  *color_frame   = gtk_hbox_new (FALSE, 0);
		     GtkWidget  *texture_frame = gtk_hbox_new (FALSE, 0);
		     GtkWidget  *color_box     = gtk_hbox_new (FALSE, 0);
		     GtkWidget  *texture_box   = gtk_vbox_new (FALSE, 0);
		     GtkWidget  *hbox;
		     proplist_t *color_data;
		     proplist_t *texture_data;
		     proplist_t *font_data;
		     
		     gtk_container_add (GTK_CONTAINER (subpage), vbox);
		     
		     /*
		      *  Overall boxes
		      */
		     gtk_box_pack_start (GTK_BOX (vbox), nw_hbox,
					 FALSE, TRUE, 5);
		     hbox = gtk_hbox_new (FALSE, 0);
		     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, TRUE, 0);
		     gtk_box_pack_start (GTK_BOX (hbox), color_frame,
					 TRUE, TRUE, 5);
		     hbox = gtk_hbox_new (FALSE, 0);
		     gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 5);
		     gtk_box_pack_start (GTK_BOX (hbox), texture_frame,
					 TRUE, TRUE, 5);
		     while (gtk_events_pending())
			gtk_main_iteration();
		     
		     /*
		      *  Color box
		      */
		     {
			GtkWidget  *frame = gtk_frame_new (_("Colors"));
			GtkWidget  *box   = color_box;
			
			gtk_box_pack_start (GTK_BOX (color_frame), frame,
					    TRUE, TRUE, 0);
			gtk_frame_set_label_align (GTK_FRAME (frame), 0.5, 0.5);
			gtk_container_add (GTK_CONTAINER (frame), box);
			gtk_container_set_border_width (GTK_CONTAINER (box), 5);
			color_data = PLMakeData ((unsigned char *) &box,
						 sizeof (GtkWidget *));
			gtk_object_set_user_data (GTK_OBJECT (box),
						  color_frame);
		     }
		     while (gtk_events_pending())
			gtk_main_iteration();
		     /*
		      *  Texture box
		      */
		     {
			GtkWidget  *frame = gtk_frame_new (_("Texture"));
			GtkWidget  *box   = texture_box;

			gtk_box_pack_start (GTK_BOX (texture_frame), frame,
					    TRUE, TRUE, 0);
			gtk_frame_set_label_align (GTK_FRAME (frame), 0.5, 0.5);
			gtk_container_add (GTK_CONTAINER (frame), box);
			texture_data = PLMakeData ((unsigned char *) &box,
						   sizeof (GtkWidget *));
			gtk_object_set_user_data (GTK_OBJECT (box),
						  texture_frame);
		     }
		     /*
		      *  Font box
		      */
		     {
			font_data = PLMakeData ((unsigned char *) &nw_hbox,
						sizeof (GtkWidget *));
		     }
		     gtk_widget_show_all (subpage);
		     gtk_widget_hide (color_frame);
		     gtk_widget_hide (texture_frame);

		     while (gtk_events_pending())
			gtk_main_iteration();
		     
		     /*
		      *  Make option box with subsubsections and subboxes
		      *  for color, texture and font
		      */
		     if (subtype && PLIsDictionary (subtype))
		     {
			proplist_t *ssentries = PLGetAllDictionaryKeys (subtype);
			unsigned   k;
			char	   *path;
			GtkWidget  *option_menu = gtk_option_menu_new ();
			GtkWidget  *menu        = gtk_menu_new ();

			path = g_strconcat (PLGetString (thispanel), "/",
					    PLGetString (thissub), NULL);

			for (k = 0; k < PLGetNumberOfElements (ssentries); k++)
			{
			   proplist_t *thisss;
			   proplist_t *ssentry;
			   proplist_t *ssinfo;
			   GtkWidget  *menu_item;
			   
			   thisss    = PLGetArrayElement (ssentries, k);
			   ssentry   = PLGetDictionaryEntry (subtype, thisss);
			   ssinfo    = PLGetDictionaryEntry (ssentry, plinfo);
			   menu_item = gtk_menu_item_new_with_label (D_(PLGetString (ssinfo)));
			   gtk_signal_connect (GTK_OBJECT (menu_item), "activate",
					       GTK_SIGNAL_FUNC (switch_subentries),
					       path);
			   gtk_object_set_user_data (GTK_OBJECT (menu_item),
						     (gpointer) PLGetString (thisss));
			   gtk_menu_append (GTK_MENU (menu), menu_item);
			   gtk_widget_show (menu_item);

			   PLInsertDictionaryEntry (ssentry, plcolor,
						    color_data);
			   PLInsertDictionaryEntry (ssentry, pltexture,
						    texture_data);
			   PLInsertDictionaryEntry (ssentry, plfont,
						    font_data);
			   if (strcaseeq (PLGetString (thisss), "Widget"))
			      widgetstyle_item = menu_item;
			}
			gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu),
						  menu);
			gtk_option_menu_set_history (GTK_OPTION_MENU (option_menu), 0);
			gtk_widget_show (option_menu);
			options = g_list_append (options, option_menu);

			while (gtk_events_pending())
			   gtk_main_iteration();
			
			{
			   GtkWidget *frame;
			   GtkWidget *box = gtk_hbox_new (TRUE, 0);

			   frame = gtk_frame_new (D_(PLGetString (subinfo)));
			   gtk_frame_set_label_align (GTK_FRAME (frame),
						      0.5, 0.5);
			   gtk_container_set_border_width (GTK_CONTAINER (box), 5);
			   gtk_container_add (GTK_CONTAINER (frame), box);
			   gtk_box_pack_start (GTK_BOX (box), option_menu,
					       TRUE, TRUE, 0);
			   gtk_box_pack_start (GTK_BOX (nw_hbox), frame,
					       FALSE, TRUE, 5);
			   gtk_widget_show_all (frame);
			}

			while (gtk_events_pending())
			   gtk_main_iteration();

		     }
		     else
		     {
			PLInsertDictionaryEntry (subentry, plcolor,
						 color_data);
			PLInsertDictionaryEntry (subentry, pltexture,
						 texture_data);
			PLInsertDictionaryEntry (subentry, plfont,
						 font_data);
		     }
		     PLRelease (color_data);
		     PLRelease (texture_data);
		     PLRelease (font_data);
		  }
	       }
	       PLRelease (subentries);
	    }
	 }
      }
      PLRelease (tab);
   }

   while (gtk_events_pending())
      gtk_main_iteration();

   /*
    *  Add a dialog window for each Window Maker attribute
    */
   {
      proplist_t *all_keys = PLGetAllDictionaryKeys (wmconfig);
      unsigned   melem     = PLGetNumberOfElements (all_keys);
      unsigned	 n;

      /*
       *  Generate a dialog widget for every attribute
       */
      for (n = 0; n < melem; n++)
      {
	 proplist_t *key    = PLGetArrayElement (all_keys, n);
	 proplist_t *value  = PLGetDictionaryEntry (windowmaker, key);
	 proplist_t *keydef = PLGetDictionaryEntry (wmconfig, key);
	 proplist_t *group  = PLGetDictionaryEntry (keydef, plgrp);
	 proplist_t *title  = PLGetDictionaryEntry (keydef, pltitle);
	 proplist_t *range;
	 proplist_t *info;
	 GtkWidget  *page_box;
	 char	    *name;
	 char	    *type;
	 proplist_t *ptype;
	 proplist_t *pentry;
	 
	 if (strcaseeq (PLGetString (group), "None"))
	    continue;
	 
	 range = PLGetDictionaryEntry (keydef, plrange);
	 info  = PLGetDictionaryEntry (keydef, plinfo);
	 name  = PLGetString (key);
	 type  = PLGetString (PLGetDictionaryEntry (keydef, pltype));

#ifndef FREEBSD_SYSTEM   
	 /*
	  *  Update of progress bar window
	  */
	 gtk_label_set_text (GTK_LABEL (progress_label), PLGetString (key));
	 gtk_progress_bar_update (GTK_PROGRESS_BAR (progress_bar),
				  n / (double) melem);
	 gtk_widget_draw (progress_window, NULL);
#endif /* not FREEBSD_SYSTEM */
	 
	 while (gtk_events_pending())
	    gtk_main_iteration();

	 /*
	  *  Get main group and corresponding container
	  */
	 {
	    char	*pos;
	    proplist_t	*pgroup;
	    proplist_t  *pbox;
	    char	*groupname = g_strdup (PLGetString (group));

	    if ((pos = strchr (groupname, '/')))
	       *pos = 0;

	    pgroup = PLMakeString (groupname);
	    pentry = PLGetDictionaryEntry (panel, pgroup);
	    pbox   = PLGetDictionaryEntry (pentry, plwidget);
	    ptype  = PLGetDictionaryEntry (pentry, pltype);
	    PLRelease (pgroup);
	    
	    if (!pentry || !pbox || !ptype)
	       continue;

	    if (PLIsDictionary (ptype)) /* get sub entry */
	    {
	       char *subname = pos + 1;

	       if ((pos = strchr (subname, '/')))
		  *pos = 0;
	       
	       pgroup = PLMakeString (subname);
	       pentry = PLGetDictionaryEntry (ptype, pgroup); 
	       ptype  = PLGetDictionaryEntry (pentry, pltype);
	       PLRelease (pgroup);

	       if (PLIsDictionary (ptype)) /* get sub sub entry */
	       {
		  pgroup = PLMakeString (pos + 1);
		  pentry = PLGetDictionaryEntry (ptype, pgroup);
		  PLRelease (pgroup);
		  if (strcaseeq (type, "Color"))
		     pbox = PLGetDictionaryEntry (pentry, plcolor);
		  else if (strcaseeq (type, "Font"))
		     pbox = PLGetDictionaryEntry (pentry, plfont);
		  else if (strcaseeq (type, "Texture"))
		     pbox = PLGetDictionaryEntry (pentry, pltexture);
		  else
		     pbox = PLGetDictionaryEntry (pentry, plwidget);
	       }
	       else
	       {
		  if (strcaseeq (type, "Color"))
		     pbox = PLGetDictionaryEntry (pentry, plcolor);
		  else if (strcaseeq (type, "Font"))
		     pbox = PLGetDictionaryEntry (pentry, plfont);
		  else if (strcaseeq (type, "Texture"))
		     pbox = PLGetDictionaryEntry (pentry, pltexture);
		  else
		     pbox = PLGetDictionaryEntry (pentry, plwidget);
	       }
	       page_box = * (GtkWidget **) PLGetDataBytes (pbox);
	       {
		  GtkWidget *frame;
		  frame = gtk_object_get_user_data (GTK_OBJECT (page_box));
		  
		  if (frame && GTK_IS_WIDGET (frame)
		      && !GTK_WIDGET_VISIBLE (frame))
		     gtk_widget_show_all (frame);
	       }
	    }
	    else
	       page_box = * (GtkWidget **) PLGetDataBytes (pbox);

	    Free (groupname);
	 }

	 while (gtk_events_pending())
	    gtk_main_iteration();
 
	 /*
	  *  Use different dialogs for different attribute types
	  */
	 if (strcaseeq ("Bool", type))
	 {
	    boolean_dialog (GTK_TABLE (page_box), key, value, tooltips, info);
	 }
	 else if (strcaseeq ("String", type)
		  && !strcaseeq ("ModifierKey", PLGetString (key)))
	 {
	    string_dialog (GTK_TABLE (page_box), key, value, range,
			   tooltips, info);
	 }
	 else if (strcaseeq ("Int", type))
	 {
	    int_dialog (GTK_TABLE (page_box), key, value, range, tooltips, info);
	 }
	 else if (strcaseeq ("Text", type))
	 {
	    text_dialog (GTK_TABLE (page_box), key, value, tooltips, info);
	 }
	 else if (strcaseeq ("Font", type) || strcaseeq ("Color", type)
		  || strcaseeq ("Texture", type)) 
	 {
	    GtkWidget	*widget;
	    proplist_t	*data;

	    if (strcaseeq ("Font", type)) 
	       widget = font_dialog (page_box, key, value, tooltips, info);
	    else if (strcaseeq ("Texture", type))
	    {
	       GtkWidget  *box     = NULL;
	       proplist_t *fontbox = PLGetDictionaryEntry (pentry, plfont);

	       if (fontbox)
		  box = * (GtkWidget **) PLGetDataBytes (fontbox);
	       widget = texture_dialog (page_box, box, key);
	    }
	    else
	       widget = color_dialog (page_box, key, value, tooltips, info,
				      title);

	    data = PLMakeData ((unsigned char *) &widget, sizeof (GtkWidget *));
	    PLInsertDictionaryEntry (keydef, plwidget, data);
	    PLRelease (data);

	    if (PLIsDictionary (ptype)) /* subsubbox */
	       gtk_widget_hide (widget);
	 }
	 else if (strcaseeq ("Key", type)) 
	 {
	    add_shortcut (page_box, key);
	 }
	 else if (strcaseeq ("Pathlist", type))
	 {
	    path_dialog (page_box, key, value, tooltips, info);
	 }
      }
      PLRelease (all_keys);
   }
   
   gtk_widget_set_usize (main_window,
			 min (gdk_screen_width () - 64, 650),
			 min (gdk_screen_height () - 64, 630));
   
   while (gtk_events_pending())
      gtk_main_iteration();
   
   toggle_save (changed = NO, NULL);
#ifndef FREEBSD_SYSTEM   
   gtk_widget_destroy (progress_window);
#endif /* not FREEBSD_SYSTEM */

   PLRelease (plinfo);
   PLRelease (pltype);
   PLRelease (plicon);
   PLRelease (plwidget);
   PLRelease (plcolor);
   PLRelease (plfont);
   PLRelease (pltexture);
   PLRelease (plrange);
   PLRelease (plgrp);
   PLRelease (pltitle);

   g_list_foreach (options, init_option_menu, NULL);
   g_list_free (options);

   while (gtk_events_pending())
      gtk_main_iteration();
   
   check_widgetstyle ();
#ifdef GETSTYLE
   get_theme_attributes ();
#endif /* GETSTYLE */
   
   while (gtk_events_pending())
      gtk_main_iteration();
   
   gtk_widget_show (main_window);
   gtk_main ();
   gtk_widget_destroy (main_window);
#ifdef TILE_OF_DAY
   cleanup_tileofday ();
#endif   
   cleanup_textures ();
#ifdef GETSTYLE
   cleanup_themes ();
#endif /* GETSTYLE */
}

void
check_widgetstyle (void)
{
   proplist_t *plnewstyle = PLMakeString ("NewStyle");
   proplist_t *value	  = PLGetDictionaryEntry (windowmaker, plnewstyle);

   gtk_widget_set_sensitive (widgetstyle_item, !convert_bool (value));
   
   PLRelease (plnewstyle);
}

char *
get_pixmap_path (const char *name)
/*
 *  Guess filename of pixmap 'name'.
 *  Cycle through list of available paths or use absolute path
 *  if name [0] == '/' or '~'.
 *
 *  Return value:
 *	string with pixmap path on success
 *	NULL if pixmap not found
 */
{
   if (name [0] == '/' || name [0] == '~') /* absolute path ? */
   {
      FILE *file;
      char *path = expand_tilde (name);

      file = fopen (path, "r");
      if (!file)
      {
	 Free (path);
	 return NULL;
      }
      else
      {
	 fclose (file);
	 return path;
      }
   }
   else
   {
      unsigned	 n;
      char	 *found	      = NULL;
      proplist_t *plpath      = PLMakeString ("PixmapPath");
      proplist_t *pixmap_path = PLGetDictionaryEntry (windowmaker, plpath);
      
      PLRelease (plpath);

      for (n = 0; n < PLGetNumberOfElements (pixmap_path) && !found; n++)
      {
	 FILE *file;
	 char *path = expand_tilde (PLGetString (PLGetArrayElement (pixmap_path,
								    n)));
	 char *filename = g_strconcat (path, "/", name, NULL);

	 Free (path);
	 file = fopen (filename, "r");
	 if (file)
	 {
	    found = filename;
	    fclose (file);
	 }
	 else
	    Free (filename);
      }
      return found;
   }
}

GtkWidget *
make_pixmap (const char *name, int width, int height, GtkWidget *pw)
/*
 *  Genearate a new pixmap widget with pixmap 'name' of maximum size
 *  'max_width' x 'max_height'. If 'pw' is already valid (i.e., != NULL)
 *  just replace the pixmap of the widget. 
 *
 *  Return value:
 *	new or changed pixmap widget
 */
{
   char *path = get_pixmap_path (name);

   if (!path)
   {
      dialog_popup (DIALOG_ERROR, NULL, NULL,
		    _("Can't find pixmap\n'%s`\nin PixmapPath. Please update\n"
		      "your PixmapPath attribute."), name);
      return make_pixmap (PKGDATADIR "/black.xpm", width, height, pw);
   }

#ifdef HAVE_STAT
   /*
    *  Don't load large XPM (may cause trouble with imlib)
    */
   {
      struct stat buffer;
      
      if (strstr (path, ".xpm") && stat (path, &buffer) == 0
	  && buffer.st_size > 65000) /* image too large */
      {
	 Free (path);
	 return make_pixmap (PKGDATADIR "/black.xpm", width, height, pw);
      }
   }
#endif /* HAVE_STAT */

#if defined(PREVIEWS) && !defined(CONVERT)

   if (!pw && width > 0 && height > 0)
      pw = make_image (PKGDATADIR "/black.xpm", width, height, pw);
   pw = make_image (path, width, height, pw);
   
#else  /* not PREVIEWS or CONVERT */
   
   {
      GdkBitmap *mask   = NULL;
      GdkPixmap *pixmap = NULL;

#ifdef HAVE_IMLIB
      if (use_imlib)
      {
	 GdkImlibImage *image;

	 image = gdk_imlib_load_image (path);
	 if (image)
	 {
	    if (width <= 0)
	       width = image->rgb_width;
	    if (height <= 0)
	       height = image->rgb_height;
	    if (image->rgb_width > width || image->rgb_height > height)
	    {
	       if (height * image->rgb_width > width * image->rgb_height)
		  height = width * image->rgb_height / image->rgb_width;
	       else
		  width  = height * image->rgb_width / image->rgb_height;
	    }
	    gdk_imlib_render (image, max (width, 22), max (height, 22));
	    pixmap = gdk_imlib_move_image (image);
	    mask   = gdk_imlib_move_mask (image);
	    if (strchr (g_basename (path), '.')) /* don't cache x-of-day */
	       gdk_imlib_destroy_image (image);
	    else
	       gdk_imlib_kill_image (image);
	 }
      }
      else
#else  /* not HAVE_IMLIB */
      {
	 pixmap
	    = gdk_pixmap_create_from_xpm (main_window->window, &mask,
					  &main_window->style->bg [GTK_STATE_NORMAL],
					  PKGDATADIR "/black.xpm"); 
      }
#endif /* not HAVE_IMLIB */
      if (!pixmap)			/* pixmap not found or unknown format */
      {
	 static bool_t error_message = NO;

	 if (!error_message)
	 {
	    dialog_popup (DIALOG_ERROR, NULL, NULL,
			  _("Unable to display pixmap\n'%s`\n"
			    "(format not supported).\n"
			    "See stderr for subsequent messages."), path);
	    error_message = YES;
	 }
	 else
	    warning (_("Can't display pixmap '%s`."), path);
	 Free (path);
	 return make_pixmap (PKGDATADIR "/black.xpm", width, height, pw);
      }
      Free (path);
      if (pw)
	 gtk_pixmap_set (GTK_PIXMAP (pw), pixmap, mask);
      else
      {
	 pw = gtk_pixmap_new (pixmap, mask);
	 gtk_widget_show (pw);
      }
   }
#endif /* not PREVIEWS or CONVERT */

   return pw;
}

gint
quit (GtkWidget *widget, gpointer ptr)
/*
 *  Return from gtk_main () loop, i.e. leave the application.
 *  Before quitting ask for confirmation.
 *
 *  No return value.
 */
{
   if (changed || rootmenu_changed ()

#if defined(HAVE_POPEN) && defined(TILE_OF_DAY) && defined(CRONTAB)   

       || crontab_changed ()

#endif /* HAVE_POPEN && TILE_OF_DAY */

       )
   {
      quit_and_save_popup (_("Some values are modified.\n"
			     "Do you want to save them before quitting?"),
			   save_config_file, (gpointer) "quit");
      return TRUE;
   }
   else
   {
      gtk_main_quit ();
      return TRUE;				/* do not delete window */
   }
}

static void
quit_and_save_popup (char *question,
		     void (*save_function) (GtkWidget *, gpointer),
		     gpointer dataptr)
/*
 *  Popup dialog to ensure that current process is not killed until
 *  user confirmation.
 *  'question' is the text to be displayed.
 *  'save_function' is called in case user confirms saving.
 *  'dataptr' argument is passed to 'save_function'.
 *
 *  No return value.
 */
{
   GtkWidget *hbox;
   GtkWidget *window, *button;
   
   window = gtk_dialog_new ();
   gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);
   gtk_window_set_title (GTK_WINDOW (window), _("Save before quitting?"));
   gtk_widget_realize (window);

   gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		       GTK_SIGNAL_FUNC (gtk_false), NULL);
   gtk_signal_connect_object (GTK_OBJECT (window), "destroy",
			      GTK_SIGNAL_FUNC (gtk_widget_destroy),
			      GTK_OBJECT (window));

   hbox = gtk_hbox_new (FALSE, 5);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->vbox), hbox,
		       FALSE, FALSE, 0);
   gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);

   gtk_box_pack_start (GTK_BOX (hbox),
		       gtk_pixmap_new (p_array [P_INFO].pixmap,
				       p_array [P_INFO].mask),
		       TRUE, FALSE, 5);
   
   gtk_box_pack_start (GTK_BOX (hbox), gtk_label_new (question), TRUE, TRUE, 5);

   button = gtk_button_new_with_label (_("Yes"));
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->action_area), button,
		       TRUE, TRUE, 10);
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
   gtk_widget_grab_default (button);
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
		       GTK_SIGNAL_FUNC (save_function), dataptr);
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			      (GtkSignalFunc) gtk_widget_destroy,
			      GTK_OBJECT (window));

   button = gtk_button_new_with_label (_("No"));
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->action_area), button,
		       TRUE, TRUE, 10);
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
		       GTK_SIGNAL_FUNC (gtk_main_quit), NULL);

   button = gtk_button_new_with_label (_("Cancel"));
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->action_area), button,
		       TRUE, TRUE, 10);
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			      (GtkSignalFunc) gtk_widget_destroy,
			      GTK_OBJECT (window));

   gtk_widget_show_all (window);
}

void
save_config_file (GtkWidget *widget, gpointer ptr)
/*
 *  Save 'WindowMaker' file if attributes have been changed.
 *
 *  No return value.
 *
 *  Side effects:
 *	'changed' is set to FALSE
 */
{
#if defined(HAVE_POPEN) && defined(TILE_OF_DAY) && defined(CRONTAB)
   
   if (!save_crontab ())
      return;
   
#endif /* HAVE_POPEN && TILE_OF_DAY && CRONTAB */

   if (!save_rootmenu ())
      return;
   
   if (changed)
   {
      proplist_t *newwm = PLMakeDictionaryFromEntries (NULL, NULL, NULL);
      
      PLSetFilename (newwm, PLRetain (PLGetFilename (windowmaker)));
      /*
       *  Compare user attributes with system attributes
       */
      {
	 proplist_t *all_keys = PLGetAllDictionaryKeys (windowmaker) ;
	 unsigned	 n;
      
	 for (n = 0; n < PLGetNumberOfElements (all_keys); n++)
	 {
	    proplist_t *user;
	    proplist_t *key = PLGetArrayElement (all_keys, n);
	 
	    user = PLGetDictionaryEntry (windowmaker, key);

	    if (PLGetDictionaryEntry (plist_changed, key))
	       PLInsertDictionaryEntry (newwm, key, PLDeepCopy (user));
	 }
	 PLRelease (all_keys);
      }

      if (PLSave (newwm, YES))
      {
	 changed = NO;
	 message (_("Window Maker config file '%s' saved."),
		  PLGetString (PLGetFilename (newwm)));
      }
      else
      {
	 dialog_popup (DIALOG_ERROR, NULL, NULL,
		       _("Can't write to Window Maker file\n`%s'."),
		       PLGetString (PLGetFilename (newwm)));
	 return;
      }
      PLRelease (newwm);
   }

   enable_revert ();
   toggle_save (FALSE, NULL);
   
   if (ptr && streq (ptr, "quit"))
      gtk_main_quit ();
}

/*******************************************************************************

			       private code
  
*******************************************************************************/

static void
init_option_menu (gpointer data, gpointer user_data)
{
   GtkOptionMenu *omenu = GTK_OPTION_MENU (data);
   GtkMenu	 *menu 	= GTK_MENU (gtk_option_menu_get_menu (omenu));
   GtkWidget	 *item 	= gtk_menu_get_active (menu);

   if (item)
      gtk_menu_item_activate (GTK_MENU_ITEM (item));
}

static GtkWidget *
create_page (GtkNotebook *notebook, const char *name, GtkWidget *pixmap)
/*
 *  Create a new group notebook page and append it to the given 'notebook'.
 *  'name' is the notebook label and 'pm' is the pixmap identifier.
 *
 *  Return value:
 *	container of label and pixmap
 */
{
   GtkWidget *vbox      = gtk_vbox_new (FALSE, 5);
   GtkWidget *label_box = gtk_vbox_new (FALSE, 5);
   
   gtk_container_set_border_width (GTK_CONTAINER (label_box), 5);

   if (pixmap)
      gtk_box_pack_start (GTK_BOX (label_box), pixmap, FALSE, TRUE, 0);
   gtk_box_pack_start (GTK_BOX (label_box), gtk_label_new (name), FALSE, TRUE, 0);
   gtk_widget_show_all (label_box);
      
   gtk_widget_show_all (vbox);
   gtk_notebook_append_page_menu (notebook, vbox, label_box, NULL);

   return vbox;
}

static void
switch_subentries (GtkWidget *widget, gpointer ptr)
/*
 *  Show only widgets of chosen subsubentry
 */
{
   unsigned	n;
   proplist_t	*all_keys = PLGetAllDictionaryKeys (wmconfig);
   proplist_t	*plgrp    = PLMakeString ("Group");
   proplist_t	*plwidget = PLMakeString ("Widget");
   char         *text     = gtk_object_get_user_data (GTK_OBJECT (widget));
   char         *prefix   = (char *) ptr;
   char		*path	  = g_strconcat (prefix, "/", text, NULL);
   
   for (n = 0; n < PLGetNumberOfElements (all_keys); n++)
   {
      proplist_t *key    = PLGetArrayElement (all_keys, n);
      proplist_t *keydef = PLGetDictionaryEntry (wmconfig, key);
      proplist_t *data   = PLGetDictionaryEntry (keydef, plwidget);

      if (data)
      {
	 proplist_t *group = PLGetDictionaryEntry (keydef, plgrp);
	 GtkWidget  *widget = * (GtkWidget **) PLGetDataBytes (data);

	 if (strcaseeq (path, PLGetString (group)))
	    gtk_widget_show (widget);
	 else
	 {
	    char *right = strrchr (path, '/');

	    /*  g_strncasecmp: only in GTK+1.1.x */
	    {
	       char *str1 = g_strdup (path);
	       char *str2 = g_strdup (PLGetString (group));

	       str1 [right - path] = 0;
	       str2 [right - path] = 0;
	       
	       if (g_strcasecmp (str1, str2) == 0)
		  gtk_widget_hide (widget);
	       Free (str1);
	       Free (str2);
	    }
	 }
      }
   }
   PLRelease (plwidget);
   PLRelease (plgrp);
}
